#  Copyright 2024 The HuggingFace Team. All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


_HEAD_TO_AUTOMODELS = {
    "feature-extraction": "IPEXModel",
    "text-generation": "IPEXModelForCausalLM",
    "text-classification": "IPEXModelForSequenceClassification",
    "token-classification": "IPEXModelForTokenClassification",
    "question-answering": "IPEXModelForQuestionAnswering",
    "fill-mask": "IPEXModelForMaskedLM",
    "image-classification": "IPEXModelForImageClassification",
    "audio-classification": "IPEXModelForAudioClassification",
}
